@@ -5,16 +5,16 @@
 # it under the terms of the GNU General Public License as published by
 # the Free Software Foundation, either version 3 of the License, or
 # (at your option) any later version.
-# 
+#
 # This program is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 # GNU General Public License for more details.
-# 
+#
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
-# Rewritten from scratch, but based on the code from gnome-schedual by:
+# Rewritten from scratch, but based on the code from gnome-schedule by:
 # - Philip Van Hoof <me at pvanhoof dot be>
 # - Gaute Hope <eg at gaute dot vetsj dot com>
 # - Kristof Vansant <de_lupus at pandora dot be>
@@ -23,7 +23,8 @@ EXAMPLE_USE = """
 from crontab import CronTab
 import sys
 
-cron = CronTab()
+# Create a new non-installed crontab
+cron = CronTab(tab='')
 job  = cron.new(command='/usr/bin/echo')
 
 job.minute.during(5,50).every(5)
@@ -56,15 +57,25 @@ for job6 in cron:
 cron.remove_all('echo')
 cron.remove_all('/foo/bar')
 cron.write()
+
+# Croniter Extentions allow you to ask for the scheduled job times, make
+# sure you have croniter installed, it's not a hard dependancy.
+
+job3.schedule().get_next()
+job3.schedule().get_prev()
+
 """
 
 import os, re, sys
 import tempfile
+import subprocess as sp
+
+from datetime import datetime
+from cronlog import CronLog
 
 __pkgname__ = 'python-crontab'
-__version__ = '1.3.2'
+__version__ = '1.4'
 
-CRONCMD = "/usr/bin/crontab"
 ITEMREX = re.compile('^\s*([^@#\s]+)\s+([^@#\s]+)\s+([^@#\s]+)' +
     '\s+([^@#\s]+)\s+([^@#\s]+)\s+([^#\n]*)(\s+#\s*([^\n]*)|$)')
 SPECREX = re.compile('@(\w+)\s([^#\n]*)(\s+#\s*([^\n]*)|$)')
@@ -99,14 +110,32 @@ S_INFO = [
     { 'name' : 'Day of Week',  'max_v' : 7,  'min_v' : 0, 'enum' : WEEK_ENUM },
 ]
 
-# Detect Python3
+# Detect Python3 and which OS for temperments.
 import platform
 py3 = platform.python_version()[0] == '3'
+WinOS = platform.system() == 'Windows'
+SunOS = not WinOS and (os.uname()[0] == "SunOS" or os.getenv('SunOS_TEST'))
+CRONCMD = "/usr/bin/crontab"
+if sys.argv[0].startswith('test_'):
+    CRONCMD = './crontest'
 
 if py3:
     unicode = str
     basestring = str
 
+try:
+    # Croniter is an optional import
+    from croniter import croniter as _croniter
+    class croniter(_croniter):
+        """Same as normal croniter, but always return datetime objects"""
+        def get_next(self, type_ref=datetime):
+            return _croniter.get_next(self, type_ref)
+        def get_prev(self, type_ref=datetime):
+            return _croniter.get_prev(self, type_ref)
+        def get_current(self, type_ref=datetime):
+            return _croniter.get_current(self, type_ref)
+except ImportError:
+    croniter = None
 
 
 class CronTab(object):
@@ -116,19 +145,26 @@ class CronTab(object):
     user    - Set the user of the crontab (defaults to $USER)
     tab     - Use a string variable as the crontab instead of installed crontab
     tabfile - Use a file for the crontab instead of installed crontab
-    compat  - Force disable some features for SunOS (automatic).
+    log     - Filename for logfile instead of /var/log/syslog
 
     """
-    def __init__(self, user=None, tab=None, tabfile=None, compat=False):
-        self.user  = user
-        self.root  = ( os.getuid() == 0 )
+    def __init__(self, user=None, tab=None, tabfile=None, log=None):
         self.lines = None
         self.crons = None
         self.filen = None
+        # Protect windows users
+        self.root  = not WinOS and os.getuid() == 0
+        self.user  = user
         # Detect older unixes and help them out.
-        self.compat = compat or os.uname()[0] == "SunOS"
         self.intab = tab
         self.read(tabfile)
+        self._log = log
+
+    @property
+    def log(self):
+        if self._log == None or isinstance(self._log, basestring):
+            self._log = CronLog(self._log, user=self.user or 'root')
+        return self._log
 
     def read(self, filename=None):
         """
@@ -138,15 +174,17 @@ class CronTab(object):
         self.crons = []
         self.lines = []
         if self.intab != None:
-          lines = self.intab.split('\n')
+            lines = self.intab.split('\n')
         elif filename:
-          self.filen = filename
-          with open(filename, 'r') as fhl:
-              lines = fhl.readlines()
+            self.filen = filename
+            with open(filename, 'r') as fhl:
+                lines = fhl.readlines()
         else:
-          lines = os.popen(self._read_execute()).readlines()
+            p = sp.Popen(self._read_execute(), stdout=sp.PIPE)
+            (out, err) = p.communicate()
+            lines = out.split("\n")
         for line in lines:
-            cron = CronItem(line, compat=self.compat)
+            cron = CronItem(line, cron=self)
             if cron.is_valid():
                 self.crons.append(cron)
                 self.lines.append(cron)
@@ -176,7 +214,7 @@ class CronTab(object):
 
         if not self.filen:
             # Add the entire crontab back to the user crontab
-            os.system(self._write_execute(path))
+            sp.Popen(self._write_execute(path)).wait()
             os.unlink(path)
 
     def render(self):
@@ -195,7 +233,7 @@ class CronTab(object):
 
         Returns the new CronItem object.
         """
-        item = CronItem(command=command, meta=comment, compat=self.compat)
+        item = CronItem(command=command, meta=comment, cron=self)
         self.crons.append(item)
         self.lines.append(item)
         return item
@@ -229,17 +267,17 @@ class CronTab(object):
 
     def _read_execute(self):
         """Returns the command line for reading a crontab"""
-        return "%s -l%s" % (CRONCMD, self._user_execute())
+        return [ CRONCMD, '-l' ] + self._user_execute()
 
     def _write_execute(self, path):
         """Return the command line for writing a crontab"""
-        return "%s %s%s" % (CRONCMD, path, self._user_execute())
+        return [ CRONCMD, path ] + self._user_execute()
 
     def _user_execute(self):
         """User command switches to append to the read and write commands."""
         if self.user:
-            return ' -u %s' % str(self.user)
-        return ''
+            return [ '-u', str(self.user) ]
+        return []
 
     def __iter__(self):
         return self.crons.__iter__()
@@ -253,15 +291,18 @@ class CronItem(object):
     An item which objectifies a single line of a crontab and
     May be considered to be a cron job object.
     """
-    def __init__(self, line=None, command='', meta='', compat=False):
+    def __init__(self, line=None, command='', meta='', cron=None):
         self.valid = False
         self.enabled = True
         self.slices  = []
         self.special = False
-        self.compat  = compat
-        self.set_slices()
+        self.cron  = cron
+
         self._meta   = meta
-        if line:
+        self._log    = None
+
+        self.set_slices()
+        if line and line.strip():
             self.parse(line.strip())
         elif command:
             self.command = CronCommand(unicode(command))
@@ -269,7 +310,7 @@ class CronItem(object):
 
     def parse(self, line):
         """Parse a cron line string and save the info as the objects."""
-        if line[0] == '#':
+        if not line or line[0] == '#':
             self.enabled = False
             line = line[1:].strip()
         result = ITEMREX.findall(line)
@@ -277,9 +318,14 @@ class CronItem(object):
             o_value = result[0]
             self.command = CronCommand(o_value[5])
             self._meta   = o_value[7]
-            self.set_slices( o_value )
-            self.valid = True
+            try:
+                self.set_slices( o_value )
+            except KeyError:
+                self.enabled = False
+            else:
+                self.valid = True
         elif line.find('@') < line.find('#') or line.find('#')==-1:
+
             result = SPECREX.findall(line)
             if result and result[0][0] in SPECIALS:
                 o_value = result[0]
@@ -299,7 +345,7 @@ class CronItem(object):
             if not o_value:
                 o_value = [None, None, None, None, None]
             self.slices.append(
-                CronSlice(compat=self.compat, value=o_value[i_value],
+                CronSlice(job=self, value=o_value[i_value],
                     **S_INFO[i_value]))
 
     def enable(self, enabled=True):
@@ -316,14 +362,13 @@ class CronItem(object):
         """Return true if this job is valid"""
         return self.valid
 
+    def render_time(self):
+        return ' '.join([ unicode(self.slices[i]) for i in range(0, 5) ])
+
     def render(self):
         """Render this set cron-job to a string"""
-        time = ''
-        if not self.special:
-            slices = []
-            for i in range(0, 5):
-                slices.append(unicode(self.slices[i]))
-            time = ' '.join(slices)
+        time = self.render_time()
+
         if self.special or time in SPECIALS.values():
             if self.special:
                 time = self.special
@@ -353,6 +398,22 @@ class CronItem(object):
         for slice_v in self.slices:
             slice_v.clear()
 
+    def schedule(self, date_from=None):
+        """Return a croniter schedule is available."""
+        if not date_from:
+            date_from = datetime.now()
+        if croniter:
+            return croniter(self.render_time(), date_from)
+        raise ImportError("Croniter is not available. Please install croniter"+\
+                         " python module via pip or your package manager")
+
+    @property
+    def log(self):
+        """Return a cron log specific for this job only"""
+        if not self._log and self.cron:
+            self._log = self.cron.log.for_program(self.command)
+        return self._log
+
     @property
     def minute(self):
         """Return the minute slice"""
@@ -396,13 +457,13 @@ class CronItem(object):
 
 class CronSlice(object):
     """Cron slice object which shows a time pattern"""
-    def __init__(self, name, min_v, max_v, enum=None, value=None, compat=False):
-        self.name   = name
-        self.min    = min_v
-        self.max    = max_v
-        self.enum   = enum
-        self.compat = compat
-        self.parts  = []
+    def __init__(self, name, min_v, max_v, enum=None, value=None, job=None):
+        self.name  = name
+        self.min   = min_v
+        self.max   = max_v
+        self.job   = job
+        self.enum  = enum
+        self.parts = []
         if value:
             self._set_value(value)
 
@@ -469,13 +530,13 @@ class CronSlice(object):
         try:
             out = CronValue(v, self.enum)
         except ValueError:
-            raise ValueError("Unrecognised value '%s' for '%s'" % (v, self.name))
+            raise ValueError("Unrecognised '%s'='%s'" % (self.name, v))
         except KeyError:
-            raise KeyError("No enumeration for '%s' found '%s'" % (self.name, v))
+            raise KeyError("No enumeration '%s' got '%s'" % (self.name, v))
 
         if int(out) < self.min and int(out) > self.max:
             raise ValueError("Invalid value '%s', expected %d-%d for %s" % (
-                str(value), self.min, self.max, self.name))
+                str(v), self.min, self.max, self.name))
         return out
 
 
@@ -525,6 +586,7 @@ class CronRange(object):
     def __init__(self, vslice, *vrange):
         self.slice = vslice
         self.seq   = 1
+        self.cron  = None
 
         if not vrange:
             self.all()
@@ -563,7 +625,7 @@ class CronRange(object):
             value = _render_values([self.vfrom, self.vto], '-', resolve)
         if self.seq != 1:
             value += "/%d" % self.seq
-        if value != '*' and self.slice.compat:
+        if value != '*' and SunOS:
             value = ','.join(map(str, range(self.vfrom, self.vto+1, self.seq)))
         return value
 
@@ -585,7 +647,7 @@ class CronRange(object):
 
 
 class CronCommand(object):
-    """Reprisent a cron command as an object."""
+    """Represent a cron command as an object."""
     def __init__(self, line):
         self._command = line
 
